"""main port-for functionality."""

import random
import socket
from itertools import chain
from typing import Iterable, Type, TypeVar

from port_for import ephemeral, utils

from ._ranges import UNASSIGNED_RANGES
from .exceptions import PortForException

SYSTEM_PORT_RANGE = (0, 1024)


def select_random(
    ports: set[int] | None = None,
    exclude_ports: Iterable[int] | None = None,
) -> int:
    """Return random unused port number."""
    if ports is None:
        ports = available_good_ports()

    if exclude_ports is None:
        exclude_ports = set()

    ports.difference_update(set(exclude_ports))

    for port in random.sample(tuple(ports), min(len(ports), 100)):
        if not port_is_used(port):
            return port
    raise PortForException("Can't select a port")


def is_available(port: int) -> bool:
    """Return if port is good to choose."""
    return port in available_ports() and not port_is_used(port)


def available_ports(
    low: int = 1024,
    high: int = 65535,
    exclude_ranges: list[tuple[int, int]] | None = None,
) -> set[int]:
    """Return a set of possible ports.

    .. note::

        Excluding system, ephemeral and well-known ports.

    Pass ``high`` and/or ``low`` to limit the port range.
    """
    if exclude_ranges is None:
        exclude_ranges = []
    available = utils.ranges_to_set(UNASSIGNED_RANGES)
    exclude = utils.ranges_to_set(
        # Motivation behind excluding ephemeral port ranges:
        # let's say you decided to use an ephemeral local port
        # as a persistent port, and "reserve" it to your software.
        # OS won't know about it, and still can try to use this port.
        # This is not a problem if your service is always running and occupying
        # this port (OS would pick next one). But if the service is temporarily
        # not using the port (because of restart of other reason),
        # OS might reuse the same port,
        # which might prevent the service from starting.
        ephemeral.port_ranges()
        + exclude_ranges
        + [SYSTEM_PORT_RANGE, (SYSTEM_PORT_RANGE[1], low), (high, 65536)]
    )
    return available.difference(exclude)


def good_port_ranges(
    ports: set[int] | None = None, min_range_len: int = 20, border: int = 3
) -> list[tuple[int, int]]:
    """Return a list of 'good' port ranges.

    Such ranges are large and don't contain ephemeral or well-known ports.
    Ranges borders are also excluded.
    """
    min_range_len += border * 2
    if ports is None:
        ports = available_ports()
    ranges = utils.to_ranges(list(ports))
    lenghts = sorted([(r[1] - r[0], r) for r in ranges], reverse=True)
    long_ranges = [length[1] for length in lenghts if length[0] >= min_range_len]
    without_borders = [(low + border, high - border) for low, high in long_ranges]
    return without_borders


def available_good_ports(min_range_len: int = 20, border: int = 3) -> set[int]:
    """List available good ports."""
    return utils.ranges_to_set(good_port_ranges(min_range_len=min_range_len, border=border))


def port_is_used(port: int, host: str = "127.0.0.1") -> bool:
    """Return if port is used.

    If we can connect to the port or we cannot bind to it, it's used.
    """
    # Used if something is listening on the port, and we can connect to it
    if _accepts_connection(port, host):
        return True
    # Used if we cannot bind to the port.
    if not _can_bind(port, host):
        return True
    return False


def _can_bind(port: int, host: str) -> bool:
    """Try binding on common addresses to detect if the port is free.

    Some platforms (notably Windows) allow binding to 127.0.0.1 even when the
    port is already bound on INADDR_ANY (0.0.0.0). To reliably detect usage, we
    attempt to bind on both the requested host and INADDR_ANY. If either bind
    fails, consider the port as used.
    """
    hosts_to_try = []
    # Always try the requested host first
    hosts_to_try.append(host)
    # Also try INADDR_ANY for IPv4 to catch cases where 0.0.0.0 is occupied
    if host not in ("", "0.0.0.0"):
        hosts_to_try.append("")
        hosts_to_try.append("0.0.0.0")

    for h in hosts_to_try:
        with socket.socket() as sock:
            try:
                sock.bind((h, port))
            except socket.error:
                return False
    return True


def _accepts_connection(port: int, host: str) -> bool:
    """Return True if connect_ex succeeds (service is listening).

    Works reliably across platforms, including Windows.
    """
    with socket.socket() as sock:
        sock.settimeout(1)
        sock.setblocking(True)
        err = sock.connect_ex((host, port))
        # Relying on ECONNREFUSED does not produce reliable results on windows,
        # which could result in
        # either ECONREFUSED (Mapped in windows to `WSAECONNREFUSED`),
        # timeout or return any other error.
        return err == 0


T = TypeVar("T")


def filter_by_type(lst: Iterable, type_of: Type[T]) -> list[T]:
    """Return a list of elements with given type."""
    return [e for e in lst if isinstance(e, type_of)]


PortType = (
    str
    | int
    | tuple[int, int]
    | set[int]
    | list[str]
    | list[int]
    | list[tuple[int, int]]
    | list[set[int]]
    | list[set[int] | tuple[int, int]]
    | list[str | int | tuple[int, int] | set[int]]
)


def get_port(
    ports: PortType | None,
    exclude_ports: Iterable[int] | None = None,
) -> int | None:
    """Retun a random available port.

    If there's only one port passed (e.g. 5000 or '5000') function
    does not check if port is available.
    If there's -1 passed as an argument, function returns None.

    :param ports:
        exact port (e.g. '8000', 8000)
        randomly selected port (None) - any random available port
        [(2000,3000)] or (2000,3000) - random available port from a given range
        [{4002,4003}] or {4002,4003} - random of 4002 or 4003 ports
        [(2000,3000), {4002,4003}] -random of given range and set
    :param exclude_ports: A set of known ports that can not be selected.
    :returns: a random free port
    :raises: ValueError
    """
    if ports == -1:
        return None
    elif not ports:
        return select_random(None, exclude_ports)

    try:
        return int(ports)  # type: ignore[arg-type]
    except TypeError:
        pass

    ports_set: set[int] = set()

    try:
        if not isinstance(ports, list):
            ports = [ports]
        ranges: set[int] = utils.ranges_to_set(filter_by_type(ports, tuple))
        nums: set[int] = set(filter_by_type(ports, int))
        sets: set[int] = set(
            chain(
                *filter_by_type(
                    ports,
                    (set, frozenset),  # type: ignore[arg-type]
                )
            )
        )
        ports_set = ports_set.union(ranges, sets, nums)
    except ValueError:
        raise PortForException(
            f"Unknown format of ports: {ports}.\n"
            'You should provide a ports range "[(4000,5000)]"'
            'or "(4000,5000)" or a comma-separated ports set'
            '"[{4000,5000,6000}]" or list of ints "[400,5000,6000,8000]"'
            'or all of them "[(20000, 30000), {48889, 50121}, 4000, 4004]"'
        )

    return select_random(ports_set, exclude_ports)



"""Command-line utility that helps with local TCP ports management.

It finds 'good' unused TCP localhost port and remembers the association.
"""

import sys

import port_for
from port_for import PortStore


def _list(store: PortStore) -> None:
    for app, port in store.bound_ports():
        sys.stdout.write("%s: %s\n" % (app, port))


def _bind(store: PortStore, app: str, port: str | None = None) -> None:
    bound_port = store.bind_port(app, port)
    sys.stdout.write("%s\n" % bound_port)


def _unbind(store: PortStore, app: str) -> None:
    store.unbind_port(app)


def main() -> None:
    """port-for executable entrypoint."""
    import argparse

    parser = argparse.ArgumentParser(
        prog="port-for",
        description=("Command-line utility to manage local TCP port associations."),
    )
    parser.add_argument(
        "name",
        nargs="?",
        metavar="NAME",
        help="Application name to bind a port for",
    )
    parser.add_argument(
        "--bind",
        "-b",
        metavar="FOO",
        dest="bind",
        help="Find and return a port for FOO; alias for 'port-for FOO'",
    )
    parser.add_argument(
        "--port",
        "-p",
        metavar="PORT",
        dest="port",
        help="Optional specific port number for the bind command",
    )
    parser.add_argument(
        "--unbind",
        "-u",
        metavar="FOO",
        dest="unbind",
        help="Remove association for FOO",
    )
    parser.add_argument(
        "--list",
        "-l",
        action="store_true",
        dest="list_",
        help="List all associated ports",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"port-for {port_for.__version__}",
    )

    args = parser.parse_args()

    store = PortStore()

    if args.name:
        _bind(store, args.name, args.port)
    elif args.bind:
        _bind(store, args.bind, args.port)
    elif args.list_:
        _list(store)
    elif args.unbind:
        _unbind(store, args.unbind)



"""Tests for port_for.select_random."""

import pytest
from pytest import MonkeyPatch

import port_for


def test_all_used(monkeypatch: MonkeyPatch) -> None:
    """Check behaviour if there are no ports to use."""
    monkeypatch.setattr(port_for.api, "port_is_used", lambda port: True)
    with pytest.raises(port_for.PortForException):
        port_for.select_random()


def test_random_port(monkeypatch: MonkeyPatch) -> None:
    """Test random ports."""
    ports = {1, 2, 3}
    used = {1: True, 2: False, 3: True}
    monkeypatch.setattr(port_for.api, "port_is_used", lambda port: used[port])

    for x in range(100):
        assert port_for.select_random(ports) == 2



"""Tests for PortStore."""

from typing import Generator

import pytest
from pytest import TempPathFactory

import port_for
from port_for import PortStore


@pytest.fixture
def port_store(
    tmp_path_factory: TempPathFactory,
) -> Generator[PortStore, None, None]:
    """Create an initialized port store."""
    store_path = tmp_path_factory.mktemp("port_store") / "port_store.cfg"
    yield PortStore(str(store_path))


def test_store(port_store: PortStore) -> None:
    """Test port store."""
    assert port_store.bound_ports() == []

    port = port_store.bind_port("foo")
    assert port
    assert port_store.bound_ports() == [("foo", port)]
    assert port == port_store.bind_port("foo")

    port2 = port_store.bind_port("aar")
    assert port != port2
    assert port_store.bound_ports() == [("foo", port), ("aar", port2)]

    port_store.unbind_port("aar")
    assert port_store.bound_ports() == [("foo", port)]


def test_rebind(port_store: PortStore) -> None:
    """Try to rebind a used port for another app."""
    port = port_store.bind_port("foo")
    with pytest.raises(port_for.PortForException):
        port_store.bind_port("baz", port)


def test_change_port(port_store: PortStore) -> None:
    """Changing app ports is not supported."""
    port = port_store.bind_port("foo")
    another_port = port_for.select_random()
    assert port != another_port
    with pytest.raises(port_for.PortForException):
        port_store.bind_port("foo", another_port)


def test_bind_unavailable(port_store: PortStore) -> None:
    """It is possible to explicitly bind currently unavailable port."""
    port = port_store.bind_port("foo", 80)
    assert port == 80
    assert port_store.bound_ports() == [("foo", 80)]


def test_bind_non_auto(port_store: PortStore) -> None:
    """It is possible to pass a port."""
    port = port_for.select_random()
    res_port = port_store.bind_port("foo", port)
    assert res_port == port



"""PortStore implementation."""

import os
from configparser import DEFAULTSECT, ConfigParser

from .api import select_random
from .exceptions import PortForException

DEFAULT_CONFIG_PATH = "/etc/port-for.conf"


class PortStore:
    """PortStore binds, reads and stores bound ports in config."""

    def __init__(self, config_filename: str = DEFAULT_CONFIG_PATH):
        """Initialize PortStore."""
        self._config = config_filename

    def bind_port(self, app: str, port: int | str | None = None) -> int:
        """Binds port to app in the config."""
        if "=" in app or ":" in app:
            raise Exception(f'invalid app name: "{app}"')

        requested_port: str | None = None
        if port is not None:
            requested_port = str(port)

        parser = self._get_parser()

        # this app already use some port; return it
        if parser.has_option(DEFAULTSECT, app):
            actual_port = parser.get(DEFAULTSECT, app)
            if requested_port is not None and requested_port != actual_port:
                msg = (
                    f"Can't bind to port {requested_port}: "
                    f"{app} is already associated with port {actual_port}"
                )
                raise PortForException(msg)
            return int(actual_port)

        # port is already used by an another app
        app_by_port = dict((v, k) for k, v in parser.items(DEFAULTSECT))
        bound_port_numbers = map(int, app_by_port.keys())

        if requested_port is None:
            requested_port = str(select_random(exclude_ports=bound_port_numbers))

        if requested_port in app_by_port:
            binding_app = app_by_port[requested_port]
            if binding_app != app:
                raise PortForException(f"Port {requested_port} is already used by {binding_app}!")

        # new app & new port
        parser.set(DEFAULTSECT, app, requested_port)
        self._save(parser)

        return int(requested_port)

    def unbind_port(self, app: str) -> None:
        """Remove port assignement to application."""
        parser = self._get_parser()
        parser.remove_option(DEFAULTSECT, app)
        self._save(parser)

    def bound_ports(self) -> list[tuple[str, int]]:
        """List all bound ports."""
        return [(app, int(port)) for app, port in self._get_parser().items(DEFAULTSECT)]

    def _ensure_config_exists(self) -> None:
        if not os.path.exists(self._config):
            with open(self._config, "w"):
                pass

    def _get_parser(self) -> ConfigParser:
        self._ensure_config_exists()
        parser = ConfigParser()
        parser.read(self._config)
        return parser

    def _save(self, parser: ConfigParser) -> None:
        with open(self._config, "w") as f:
            parser.write(f)



"""Port for utils."""

import itertools
from typing import Iterable, Iterator


def ranges_to_set(lst: Iterable[tuple[int, int]]) -> set[int]:
    """Convert a list of ranges to a set of numbers.

    >>> ranges = [(1,3), (5,6)]
    >>> sorted(list(ranges_to_set(ranges)))
    [1, 2, 3, 5, 6]

    """
    return set(itertools.chain(*(range(x[0], x[1] + 1) for x in lst)))


def to_ranges(lst: Iterable[int]) -> Iterator[tuple[int, int]]:
    """Convert a list of numbers to a list of ranges.

    >>> numbers = [1,2,3,5,6]
    >>> list(to_ranges(numbers))
    [(1, 3), (5, 6)]

    """
    for a, b in itertools.groupby(enumerate(lst), lambda t: t[1] - t[0]):
        c = list(b)
        yield c[0][1], c[-1][1]



# -*- coding: utf-8 -*-
"""
This module/script is for updating port_for._ranges with recent information
from IANA and Wikipedia.
"""

import sys
import os
import re
import datetime
from urllib.request import Request, urlopen
from xml.etree import ElementTree
from typing import Iterator, Iterable

from port_for.utils import to_ranges, ranges_to_set

name = os.path.abspath(os.path.normpath(os.path.join(os.path.dirname(__file__), "..")))
sys.path.insert(0, name)

IANA_DOWNLOAD_URL = (
    "https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.xml"
)
IANA_NS = "http://www.iana.org/assignments"
WIKIPEDIA_PAGE = "http://en.wikipedia.org/wiki/List_of_TCP_and_UDP_port_numbers"


def _write_unassigned_ranges(out_filename: str) -> None:
    """
    Downloads ports data from IANA & Wikipedia and converts
    it to a python module. This function is used to generate _ranges.py.
    """
    with open(out_filename, "wt") as f:
        f.write("# auto-generated by port_for._download_ranges (%s)\n" % datetime.date.today())
        f.write("UNASSIGNED_RANGES = [\n")
        for range in to_ranges(sorted(list(_unassigned_ports()))):
            f.write("    (%d, %d),\n" % range)
        f.write("]\n")


def _unassigned_ports() -> set[int]:
    """Return a set of all unassigned ports (according to IANA and Wikipedia)"""
    free_ports = ranges_to_set(_parse_ranges(_iana_unassigned_port_ranges()))
    known_ports = ranges_to_set(_wikipedia_known_port_ranges())
    return free_ports.difference(known_ports)


def _wikipedia_known_port_ranges() -> Iterator[tuple[int, int]]:
    """
    Returns used port ranges according to Wikipedia page.
    This page contains unofficial well-known ports.
    """
    req = Request(WIKIPEDIA_PAGE, headers={"User-Agent": "Magic Browser"})
    page = urlopen(req).read().decode("utf8")

    # just find all numbers in table cells
    ports = re.findall(r"<td>((\d+)(\W(\d+))?)</td>", page, re.U)
    return ((int(p[1]), int(p[3] if p[3] else p[1])) for p in ports)


def _iana_unassigned_port_ranges() -> Iterator[str]:
    """
    Returns unassigned port ranges according to IANA.
    """
    page = urlopen(IANA_DOWNLOAD_URL).read()
    xml = ElementTree.fromstring(page)
    records = xml.findall("{%s}record" % IANA_NS)
    for record in records:
        description_el = record.find("{%s}description" % IANA_NS)
        assert description_el is not None
        description = description_el.text
        if description == "Unassigned":
            number_el = record.find("{%s}number" % IANA_NS)
            assert number_el is not None
            numbers = number_el.text
            assert numbers is not None
            yield numbers


def _parse_ranges(ranges: Iterable[str]) -> Iterator[tuple[int, int]]:
    """Converts a list of string ranges to a list of [low, high] tuples."""
    for txt in ranges:
        if "-" in txt:
            low, high = txt.split("-")
        else:
            low, high = txt, txt
        yield int(low), int(high)


if __name__ == "__main__":
    _write_unassigned_ranges("_ranges.py")



"""Test cases."""

import socket
import sys

import pytest

import port_for
from port_for.api import get_port
from port_for.utils import ranges_to_set


def test_common_ports() -> None:
    """Check common ports (not available)."""
    assert not port_for.is_available(80)
    assert not port_for.is_available(11211)


def test_good_port_ranges() -> None:
    """Select good ranges of ports aout of provided."""
    ranges = [
        (10, 15),  # too short
        (100, 200),  # good
        (220, 245),  # a bit short
        (300, 330),  # good
        (440, 495),  # also good
    ]

    ports = ranges_to_set(ranges)
    good_ranges = port_for.good_port_ranges(ports, 20, 3)
    assert good_ranges == [(103, 197), (443, 492), (303, 327)], good_ranges


def test_something_works() -> None:
    """Test default behaviour of good_port_ranges and available_good_ports."""
    assert len(port_for.good_port_ranges()) > 10
    assert len(port_for.available_good_ports()) > 1000


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows runner seems to allow binding low ports.",
)
def test_binding() -> None:
    """Low ports are not available (for user without root privileges)."""
    assert port_for.port_is_used(10)


def test_binding_high() -> None:
    """Test ports that are not used."""
    with socket.socket() as s:
        s.bind(("", 1025))
        port = s.getsockname()[1]
        assert port_for.port_is_used(port)
    assert not port_for.port_is_used(port)


def test_get_port_random() -> None:
    """Test case allowing get port to randomly select any port."""
    assert get_port(None)


def test_get_port_none() -> None:
    """Test special case for get_port to return None."""
    assert not get_port(-1)


def test_get_port_exclude() -> None:
    """Only one port is available at that range."""
    assert 8002 == get_port(
        (8000, 8010),
        [8000, 8001, 8003, 8004, 8005, 8006, 8007, 8008, 8009, 8010],
    )


@pytest.mark.parametrize("port", (1234, "1234"))
def test_get_port_specific(port: str | int) -> None:
    """Test special case for get_port to return same value."""
    assert get_port(port) == 1234


@pytest.mark.parametrize(
    "port_range",
    (
        [(2000, 3000)],
        (2000, 3000),
    ),
)
def test_get_port_from_range(
    port_range: list[tuple[int, int]] | tuple[int, int],
) -> None:
    """Test getting random port from given range."""
    assert get_port(port_range) in list(range(2000, 3000 + 1))


@pytest.mark.parametrize(
    "port_set",
    (
        [{4001, 4002, 4003}],
        {4001, 4002, 4003},
    ),
)
def test_get_port_from_set(port_set: list[set[int]] | set[int]) -> None:
    """Test getting random port from given set."""
    assert get_port(port_set) in {4001, 4002, 4003}


def test_port_mix() -> None:
    """Test getting random port from given set and range."""
    sets_and_ranges: list[tuple[int, int] | set[int]] = [
        (2000, 3000),
        {4001, 4002, 4003},
    ]
    want_set = set(range(2000, 3000 + 1)) | {
        4001,
        4002,
        4003,
    }
    assert get_port(sets_and_ranges) in want_set
